#!/usr/bin/python

################################################################
#
# Copyright (C) 2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

################################################################
#
# revmap.py -- virtual revision map.
#

import tlatools
import os

opt_deep_survey = 0

class revmap:

    def __init__(self, file):

        self.b_updated = 0
        self.b_dsc = 0
        self.load(file)

    def load(self, file):

        self.revmap = []
        if os.access(file, os.F_OK):
            f = open(file)

            l = f.readline()
            if l[:4] == '*DSC':
                self.b_dsc = 1
            else:
                self.b_dsc = 0

            l = f.readline()
            while l:
                self.revmap.append(l[:-1])
                l = f.readline()
            f.close()

    def save(self, file):

        f = open(file, 'w')
        if self.b_dsc == 1:
            f.write('*DSC\n')
        else:
            f.write('*SSC\n')

        for i in range(len(self.revmap)):
            f.write(self.revmap[i])
            f.write('\n')
        f.close()

    def getsize(self):

        return len(self.revmap)

    def getitem(self, n):

        return self.revmap[n]

    def finditem(self, acbvr):

        for i in range(len(self.revmap)):
            if self.revmap[i] == acbvr:
                return i

        return -1   # not found.

    def is_updated(self):

        if self.b_updated != 0:
            return 1
        else:
            return 0

    def is_dsm(self):

        if self.b_dsc == 1:
            return 1
        else:
            return 0

    def update(self, a, cbv):

        global opt_deep_survey

        if self.b_updated != 0:
            return 0       # no need for update.

        if len(self.revmap) == 0:
            revline = tlatools.get_ancestry2(a, cbv, opt_deep_survey, '')
        else:
            revline = tlatools.get_ancestry2(a, cbv, opt_deep_survey, self.revmap[-1])
            if len(revline) == 0:
                return 0   # no need to update

        if len(self.revmap) == 0:
            if len(revline) > 0:
                self.revmap = revline
                self.b_updated = 1
                self.b_dsc = opt_deep_survey
                return 1   # update
            else:
                self.b_updated = 1
                return 0   # no need for update

        newest = self.revmap[-1]

        for i in range(len(revline) - 1, -1, -1):
            if newest == revline[ i ]:
                if i == len(revline) - 1:
                    self.b_updated = 1
                    return 0 # no need for update
                else:
                    self.revmap = self.revmap + revline[ i + 1 : ]
                    self.b_updated = 1
                    return 1 # update

        return -1  # error

def _main():

    print "nop."

if __name__ == '__main__': _main()

# arch-tag: 29b43bff-0913-46ba-86d1-7465e3fe6b19
