#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#

import getopt
import string
import os
import re
import copy
import sys
import pickle
import pprint

import canvas

opt_quiet = 0
opt_reverse = 0

g_columns = 80

g_x_sep = 20
g_y_sep = 4
#b_draw_simple = 1
b_draw_simple = 0

g_script_name = ''

#
# 0 ... mine <- yours only.
# 1 ... mine -> yours only.
# 2 ... both.
#
cp_draw_mode = 2

def msg(s):

    if opt_quiet == 0:
        sys.stderr.write(s)

def gopen(fname):

    m = re.match(r"^([^|]+)\|\s*$", fname)
    if m != None:
        fname = m.group(1)
        return os.popen(fname)
    else:
        return open(fname, 'r')

def link_both(dic1, dic2, item1, item2):

    if not dic1.has_key(item1):
        dic1[item1] = []
    if dic1[item1].count(item2) == 0:
        dic1[item1].append(item2)

    if not dic2.has_key(item2):
        dic2[item2] = []
    if dic2[item2].count(item1) == 0:
        dic2[item2].append(item1)

def ntoa(n):

    div = n / 26
    mod = n % 26

    if div == 0:
        return chr(ord('A') + n)
    else:
        return chr(ord('A') + div - 1) + chr(ord('A') + mod)


def rev_append(p, n):

    p_prefix = p[ : p.rfind('-') ]
    p_n_n = p[ p.rfind('-') + 1 : ]
    dot_dot = p_n_n.rfind('..')
    if dot_dot >= 0:
        p_n1 = int( p_n_n[ : dot_dot ])
        p_n2 = int( p_n_n[ dot_dot + 2 : ] )
    else:
        p_n1 = p_n2 = int(p_n_n)

    n_prefix = n[ : n.rfind('-') ]
    n_n = int(n[ n.rfind('-') + 1 : ])

    if n_prefix == p_prefix and n_n == p_n2 + 1:
        return p_prefix + '-' + str(p_n1) + '..' + str(n_n)
    else:
        return ''

def contract_list(cl, sort=0):

    if sort:
        cl.sort(rev_cmp)

    i = 0
    while i < len(cl):

        if i == 0:
            i = i + 1
            continue

        p = cl[i - 1]
        n = cl[i]

        pn = rev_append(p, n)
        if pn != '':
            cl[i-1] = pn
            cl.pop(i)
        else:
            i = i + 1

def rev_cmp(r1, r2):

    idx1 = r1.rfind("-")
    r1_prefix = r1[:idx1]
    n1 = int(r1[idx1+1:])

    idx2 = r2.rfind("-")
    r2_prefix = r2[:idx2]
    n2 = int(r2[idx2+1:])

    return cmp([r1_prefix, n1], [r2_prefix, n2])


class Vdic:

    serialno = 0
    dic = {}
    rdic = {}

    def __init__(self):
        pass

    def add(self, rev):

        rindex = rev.rfind("--")
        if rindex == -1: # e.g. "???" case.
            return

        ver_str = rev[:rindex]

        if Vdic.dic.has_key(ver_str) == 0:
            Vdic.dic[ver_str] = ntoa(Vdic.serialno)
            Vdic.rdic[ntoa(Vdic.serialno)] = ver_str
            Vdic.serialno = Vdic.serialno + 1


    def cnv(self, rev):

        rindex = rev.rfind("--")
        ver_str = rev[:rindex]
        rev_part = rev[rindex+2:]

        return Vdic.dic[ ver_str ] + "--" + rev_part

    def rcnv(self, rev):

        rindex = rev.rfind("--")
        ver_str = rev[:rindex]
        rev_part = rev[rindex+2:]

        return Vdic.rdic[ ver_str ] + "--" + rev_part

    def __str__(self):

        for i in Vdic.dic.keys():
            print "\t", i, "\t", Vdic.dic[i]

    def emit(self):

        m = {}
        for i in Vdic.dic.keys():
            m[Vdic.dic[i]] = i

        print '''
/*
LEGEND-START'''

        mm = m.keys()
        mm.sort()

        for i in mm:
            print "\t", i, "\t", m[i]

        print '''LEGEND-END
*/'''

#    Below lines are neat, but only supported >= 2.2
#    Take compatibility.
#
#    try:
#        add = staticmethod(add)
#        cnv = staticmethod(cnv)
#        rcnv = staticmethod(rcnv)
#        emit = staticmethod(emit)
#    except NameError:
#        sys.stderr.write("Sorry, this program requires Python 2.2 or later.\n")
#        sys.exit()
    

class RevMap:

    def __init__(self):

        self.next = {}    # Key is next of Values
        self.prev = {}    # Key is prev of Values
        self.cp   = {}    # Key cherry picks Values
        self.cpby = {}    # Key is cherry picked by Values

        self.class_item = {}   # for contraction phase
        self.item_class = {}  # for contraction phase

        # singular points.

        self.roots = [None, None, None]  # common/mine/yours
        self.ends = [None, None, None]   # common/mine/yours

        '''
                                        p1       <==== self.roots[0]
                                        v
                                        p2
                                        v
                                        p3       <==== self.ends[0]
                                        |
               +------------------------+----------------+
               v                                         v
               p4   <==== self.roots[1]                  p7     <==== self.roots[2]
               v                                         v
               p5                                        p8
               v                                         v
               p6   <==== self.ends[1]                   p9     <==== self.ends[2]

        '''

        # geometry infomation.

        self.geometry = {}                    # [revision, x, y, x2, y2]
        self.geometry_scpi = [ {}, {}, {} ]   # common/mine/yours.

    def load_ancestry_graph(self, fname2):

        if len(fname2) != 2:
            raise ValueError

        dummyVdic = Vdic()

        for fi in range(len(fname2)):

            '''
                read output of "tla ancestry-graph -m ..."
                into variable "tmp".

                Note:
                    tla ancestry-graph emits the history from
                       new revisions to old revisions.
            '''
            f = gopen(fname2[fi])
            fr = f.readline()
            tmp = []
            while fr:
                fr = fr[:-1]
                m = re.match(r"^(\S+)\s+(\S+)", fr)
                if m != None:
                    m1 = m.group(1)
                    m2 = m.group(2)
                    dummyVdic.add(m1)
                    dummyVdic.add(m2)
                    tmp.append([m1, m2])
                fr = f.readline()
            f.close()

            next = self.next
            prev = self.prev
            cp = self.cp
            cpby = self.cpby

            for i in range(len(tmp) - 1):

                [b1, c1] = [tmp[i  ][0], tmp[i  ][1]]
                [b2, c2] = [tmp[i+1][0], tmp[i+1][1]]
    
                if i == 0:
                    self.ends[fi + 1] = b1
    
                if b1 != b2:
                    link_both(prev, next, b1, b2)
    
                if c1 != "???" and b1 != c1:
                    link_both(cp, cpby, b1, c1)
    
                if i == len(tmp) - 2:
                    if c2 != "???" and b2 != c2:
                        link_both(cp, cpby, b2, c2)
      
                    if self.roots[0] == None:
                        self.roots[0] = b2
                    elif self.roots[0] != b2:
                        print "These two lines don't have common ancestor. exit"
                        sys.exit(0)
    
        ''' setup singular points information '''

        b_has_forked_point = 0
        p = self.roots[0]
        while 1:
            if not self.next.has_key(p):
                break

            if len(self.next[p]) > 1:
                self.ends[0] = p
                self.roots[1] = self.next[p][0]
                self.roots[2] = self.next[p][1]
                b_has_forked_point = 1
                break
            else:
                p = self.next[p][0]

        ''' if there are no forked point,
            reset mine/yours singular point infomations.
            in this case, we have only common part of list.
        '''
        if b_has_forked_point == 0:
            self.roots[1] = self.roots[2] = None
            self.ends[1] = self.ends[2] = None
            self.ends[0] = p

    def can_contract(self, p, n):

        '''can contract or not.

        if it can, return 1.
        '''

        # 0. check lexical relationship

        [p_idx, n_idx] = [p.rfind('-'), n.rfind('-')]
        if p_idx == -1 or n_idx == -1:
            return 0
        [p_head, n_head] = [ p[:p_idx], n[:n_idx] ]
        [p_int, n_int] = [ int(p[p_idx+1:]), int(n[n_idx+1:]) ]
        if p_head != n_head or (p_int + 1) != n_int:
            return 0

        # 1. check B-edge's relationship.

        if len(self.next[p]) > 1 or len(self.prev[n]) > 1:
            return 0

        if [self.next[p][0], self.prev[n][0]] != [n, p]:
            return 0

        # 2. check C-edge's relationship.

        p_cp = self.cp.get(p, [])
        n_cp = self.cp.get(n, [])
        p_cpby = self.cpby.get(p, [])
        n_cpby = self.cpby.get(n, [])

        map(lambda x: x.sort(), [p_cp, n_cp, p_cpby, n_cpby])

        return [p_cp, p_cpby] == [n_cp, n_cpby]

    def classify_2(self, p, n, b_same):

        if not self.item_class.has_key(p):
            self.item_class[p] = p
            self.class_item[p] = [p]

        if b_same:
            p_class = self.item_class[p]
            if not self.item_class.has_key(n):
                self.item_class[n] = p_class
            if self.class_item[p_class].count(n) == 0:
                self.class_item[p_class].append(n)

        else:
            if not self.item_class.has_key(n):
                self.item_class[n] = n 
                self.class_item[n] = [n]

    def get_next_obj(self, type, b_class, r):

        '''
        get next object.
        if r == None, get first object.
        if b_class == 1, return next class instead of revision.
        type is 0(common), 1(mine), 2(yours)
        '''

        if r == None:
            return self.roots[type]

        else:
            if b_class == 0:
                if r == self.ends[type]:
                    return None
                else:
                    return self.next[r][0]
            else:
                if r == self.ends[type]:
                    return None
                r2 = self.next[r][0]
                while 1:
                    if self.item_class[r] != self.item_class[r2]:
                        return r2
                    if r2 == self.ends[type]:
                        return None
                    r2 = self.next[r2][0]

    def get_class_name(self, r):

        r_class_key = self.item_class[r]
        elements = self.class_item[r_class_key]

        if len(elements) == 0:
            return "(null)"

        if len(elements) == 1:
            return elements[0]

        elements.sort(rev_cmp)
        r_head = elements[0]
        r_tail = elements[-1]

        prefix = r_head[ : r_head.rfind("-")]
        n_1 = r_head[ r_head.rfind("-") + 1 : ]
        n_2 = r_tail[ r_tail.rfind("-") + 1 : ]

        return prefix + "-" + n_1 + ".." + n_2

    def collect_equivalence_class(self):

        for i in range(3):
            p = self.get_next_obj(i, 0, None)
            while 1:
                if p == None:
                    break
                n = self.get_next_obj(i, 0, p)
                if n == None:
                    break

                b_same = self.can_contract(p, n)
                self.classify_2(p, n, b_same)
                p = n

    def get_cherry_picked_item_list(self, rev, b_solitary, use_vdic):

        dummyVdic = Vdic()

        ret_list = []
        if self.cp.has_key(rev):
            if b_solitary:
                ret_list = filter(lambda x: not(self.next.has_key(x) or self.prev.has_key(x)), self.cp[rev])
            else:
                ret_list = filter(lambda x: self.next.has_key(x) or self.prev.has_key(x), self.cp[rev])

            if use_vdic:
                ret_list = map(lambda x: dummyVdic.cnv(x), ret_list)

        return ret_list

    def get_scpi_max_len(self, rev, use_vdic):

        cp_box_list = self.get_cherry_picked_item_list(rev, 1, use_vdic)
        contract_list(cp_box_list, 1)
        max_len = 0
        for i in range(len(cp_box_list)):
            max_len = max(max_len, len(cp_box_list[i]))

        return max_len

    def setup_geometry(self, b_class, use_vdic):

        dummyVdic = Vdic()

        #
        # max_len[0(common)/1(mine)/2(yours)][0(main)/1(scpi)]
        #     
        max_len = [[0, 0], [0, 0], [0, 0]]

        ''' get maximum revision length of each section. '''
        for i in range(3):
            p = self.get_next_obj(i, b_class, None)
            while p != None:
                if b_class:
                    p_name = self.get_class_name(p)
                else:
                    p_name = p
                if use_vdic:
                    p_name = dummyVdic.cnv(p_name)
                max_len[i][0] = max(max_len[i][0], len(p_name))
                max_len[i][1] = max(max_len[i][1], self.get_scpi_max_len(p, use_vdic))
                p = self.get_next_obj(i, b_class, p)

        ''' calculate each section's left coordinate '''

        #
        # left[0(common)/1(mine)/2(yours)][0(main)/1(scpi)]
        #     
        left = [[0, 0], [0, 0], [0, 0]]

        left_mergin = 2
        min_x_sep = 5
        rect_width = 2

        left[0][1] = left_mergin
        left[1][1] = left_mergin

        left[1][0] = left[1][1] + max_len[1][1] + rect_width + min_x_sep
        left[1][0] = max(left[1][0], g_columns * 1 / 4)

        left[2][0] = left[1][0] + max_len[1][0] + rect_width + min_x_sep * 2
        left[2][0] = max(left[2][0], g_columns * 3 / 4 - max_len[2][0] - rect_width)

        left[2][1] = left[2][0] + max_len[2][0] + rect_width + min_x_sep
        left[2][1] = max(left[2][1], g_columns - max_len[2][1] - rect_width)

        left[0][0] = left[0][1] + max_len[0][1] + rect_width + min_x_sep
        left[0][0] = max(left[0][0], (left[2][0] + max_len[2][0] + rect_width + left[1][0]) / 2 - max_len[0][0] / 2)

        ''' set each revision's coordinates. '''
        y = 0
        y_save = 0
        for i in range(3):   # common section.

            if i in [1, 2]:  # mine/yours
                y = y_save

            p = self.get_next_obj(i, b_class, None)
            while p != None:
                if b_class:
                    p_name = self.get_class_name(p)
                else:
                    p_name = p
                self.geometry[p_name] = [left[i][0], y, left[i][0] + max_len[i][0] + 1, y + 2]

                ''' calculate cherry picked coordinates. '''
                cp_box_list = self.get_cherry_picked_item_list(p, 1, use_vdic)
                contract_list(cp_box_list, 1)
                yy = y
                for cp_i in range(len(cp_box_list)):
                    self.geometry_scpi[i][ cp_box_list[cp_i] ] = [left[i][1], yy, left[i][1] + max_len[i][1] + 1, yy + 2]
                    yy = yy + g_y_sep

                y = max(yy, y + g_y_sep)

                p = self.get_next_obj(i, b_class, p)

            if i == 0:
                y = y + 4          # for fork arrow.
                y_save = y

    def draw_scp_edges_and_boxes(self, c, rev, bx1, by1, bx2, by2, type, use_vdic):

        cp_box_list = self.get_cherry_picked_item_list(rev, 1, use_vdic)
        contract_list(cp_box_list, 1)

        dummyVdic = Vdic()

        for j in range(len(cp_box_list)):
            [x1, y1, x2, y2] = self.geometry_scpi[type][ cp_box_list[j] ]
            if use_vdic:
                orig_name = dummyVdic.rcnv(cp_box_list[j])
            else:
                orig_name = cp_box_list[j]
            c.rich_center_text(x1, y1, x2, y2, cp_box_list[j], orig_name)

        if len(cp_box_list) == 1:
            if type in [0, 1]:   #common/mine
                c.line(x2+1, y1+1, bx1-1, by1, '>')
            else:
                c.line(x1-1, y1+1, bx2+1, by1, '<')
        elif len(cp_box_list) > 1:
            slist0 = map(lambda x: self.geometry_scpi[type][x], cp_box_list)
            if type in [0, 1]:
                slist = map(lambda x: [x[2]+1, x[1]+1], slist0)
                c.mline2(slist, None, [bx1 - 1, by1], '>', 2)
            else:
                slist = map(lambda x: [x[0]-1, x[1]+1], slist0)
                c.mline2(slist, None, [bx2 + 1, by1], '<', 2)

    def draw_ccp_edges(self, c, p, b_class, type):

        if b_class:
            p_name = self.get_class_name(p)
        else:
            p_name = p

        [end_x1, end_y1, end_x2, end_y2] = self.geometry[p_name]

        cpitems = self.get_cherry_picked_item_list(p, 0, 0)

        for i in range(len(cpitems)):

            #
            # FIXME: this routine has many duplicated edges.
            #

            if b_class:
                p_name2 = self.get_class_name(cpitems[i])
            else:
                p_name2 = cpitems[i]

            [start_x1, start_y1, start_x2, start_y2] = self.geometry[ p_name2 ]

            if b_draw_simple:
                if type == 1 and cp_draw_mode in [0, 2]: #mine
                    c.line(start_x1-1, start_y1+2, end_x2+1, end_y1, '<')
                if type == 2 and cp_draw_mode in [1, 2]: #yours
                    c.line(start_x2+1, start_y1+2, end_x1-1, end_y1, '>')
            else:
                if type == 1 and cp_draw_mode in [0, 2]: #mine
                    c.mline(start_x1-1, start_y1+2, end_x2+1, end_y1, '<')
                if type == 2 and cp_draw_mode in [1, 2]: #yours
                    c.mline(start_x2+1, start_y1+2, end_x1-1, end_y1, '>')

    def emit(self, b_class, use_vdic):

        dummyVdic = Vdic()

        self.setup_geometry(b_class, use_vdic)

        c = canvas.canvas(script_name = g_script_name)

        fork_info = [None, None, None]

        ''' draw boxes & base edges '''
        prev_y = None
        for i in [0]:
            p = self.get_next_obj(i, b_class, None)
            while p != None:
                if b_class:
                    p_key_name = self.get_class_name(p)
                else:
                    p_key_name = p

                if use_vdic:
                    p_disp_name = dummyVdic.cnv(p_key_name)
                else:
                    p_disp_name = p_key_name

                [x1, y1, x2, y2] = self.geometry[p_key_name]

                c.rich_center_text(x1, y1, x2, y2, p_disp_name, p_key_name)

                ''' draw cherry picked box '''
                self.draw_scp_edges_and_boxes(c, p, x1, y1, x2, y2, i, use_vdic)

                ''' draw b-edge '''
                if prev_y != None:
                    c.line((x1+x2)/2, prev_y+1, (x1+x2)/2, y1-1, 'V')
                prev_y = y2

                p = self.get_next_obj(i, b_class, p)

        fork_info[0] = [ (x1 + x2) / 2, y2 + 1]

        for i in [1, 2]:
            b_first = 1
            prev_y = None
            p = self.get_next_obj(i, b_class, None)
            while p != None:

                ''' draw main box '''

                if b_class:
                    p_key_name = self.get_class_name(p)
                else:
                    p_key_name = p

                if use_vdic:
                    p_disp_name = dummyVdic.cnv(p_key_name)
                else:
                    p_disp_name = p_key_name

                [x1, y1, x2, y2] = self.geometry[p_key_name]
                c.rich_center_text(x1, y1, x2, y2, p_disp_name, p_key_name)

                ''' draw cherry picked box '''
                self.draw_scp_edges_and_boxes(c, p, x1, y1, x2, y2, i, use_vdic)

                ''' draw b-edge '''
                if prev_y != None:
                    c.line((x1+x2)/2, prev_y+1, (x1+x2)/2, y1-1, 'V')
                prev_y = y2

                p = self.get_next_obj(i, b_class, p)

                if b_first == 1:
                    fork_info[i] = [ (x1 + x2) / 2, y1 - 1 ]
                    b_first = 0


        if fork_info[1] != None and fork_info[2] != None:
            c.draw_forked_arrow(fork_info[0], fork_info[1], fork_info[2])

        ''' draw cherry picked edges '''
        for i in [1, 2]:  # mine/yours
            p = self.get_next_obj(i, b_class, None)
            while p != None:
                self.draw_ccp_edges(c, p, b_class, i)

                p = self.get_next_obj(i, b_class, p)

        c.emit()

    def debug_draw_devline(self, type, use_vdic):

        dummyVdic = Vdic()

        r = self.get_next_obj(type, 1, None)
        y = 0
        while r != None:
            r_name = self.get_class_name(r)
            if use_vdic:
                r_name = dummyVdic.cnv(r_name)
            print r_name,
            if self.cp.has_key(r):
                cp1 = self.get_cherry_picked_item_list(r, 0, use_vdic)
                cp2 = self.get_cherry_picked_item_list(r, 1, use_vdic)

                contract_list(cp1, 1)
                contract_list(cp2, 1)
                print "   <---   ", cp1, cp2
            else:
                print ""
            r = self.get_next_obj(type, 1, r)

    def dump0(self):

        pp = pprint.PrettyPrinter(indent=4)
        print "===================== SELF.NEXT ======================="
        pp.pprint(self.next)
        print "===================== SELF.PREV ======================="
        pp.pprint(self.prev)
        print "===================== SELF.CP ======================="
        pp.pprint(self.cp)
        print "===================== SELF.CPBY ======================="
        pp.pprint(self.cpby)
        print "===================== SELF.CLASS_ITEM ======================="
        pp.pprint(self.class_item)
        print "===================== SELF.ITEM_CLASS ======================="
        pp.pprint(self.item_class)

    def dump1(self, use_vdic):

        print "------------------ COMMON -------------------------"
        self.debug_draw_devline(0, use_vdic)
        print "------------------ MINE  --------------------------"
        self.debug_draw_devline(1, use_vdic)
        print "------------------ YOURS --------------------------"
        self.debug_draw_devline(2, use_vdic)

def usage():

    sys.stderr.write(\
'''
Usage: tladdar.py [-hrS] [-H url] [-c N] fqdn-mine fqdn-yours

    -h, --help                  ... this help
    -r, --reverse               ... draw reverse ordered graph [ not yet ]
    -c, --columns=N             ... set columns number to N
    -S, --no-symbolic-converter ... do not use auto symbolic converter
    -H, --html-mode=URL         ... embed URL links in the output graph.
    -d, --debug                 ... consider fqdn-mine and fqdn-yours as file names
                                    generated by "tla ancestry-graph -m ...".

     fqdn-mine    ... just like "lord@emf.net--2003b/tla--devo--1.1"
     fqdn-yours   ... just like "stig@brautaset.org--arwen-03/file-utils--stig--1.0"

    This program caluculates twin revision topology,
    and emits the result as various formats to stdout.

''')

def tladdar(args):

    sys.argv = string.split(args, ' ')
    print sys.argv
    _main()

def _main():

#    if sys.version < "2.2":
#        sys.stderr.write("Sorry, this program requires Python 2.2 or later.\n")
#        sys.exit()

    if len(sys.argv) == 1:
        sys.stderr.write("Try 'tladdar.py -h'.\n")
        sys.exit()

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hrc:SH:d", \
                 [
                   "help",
                   "reverse",
                   "columns",
                   "no-symbolic-converter",
                   "html-mode",
                   "debug"
                 ])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    global opt_reverse
    global g_columns
    global g_script_name

    opt_reverse = 0
    opt_use_symbolic_converter = 1
    g_script_name = ''
    opt_debug = 0

    try:
        columns = os.getenv('COLUMNS')
        if columns != None:
            g_columns = int(columns)   # for CR/LF(XXX:)
    except Exception:   # old Python doesn't have "getenv" attr.
        pass

    ###########################################################################
    # parse options
    ###########################################################################
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-r", "--reverse"):
            opt_reverse = 1
        if o in ("-c", "--columns"):
            g_columns = int(a)
        if o in ("-S", "--no-symbolic-converter"):
            opt_use_symbolic_converter = 0
        if o in ("-H", "--html-mode"):
            if a[-1] == '/':
                a = a[:-1]
            g_script_name = a
        if o in ("-d", "--debug"):
            opt_debug = 1

    if len(args) > 2:
        print "Too many fqdns. exit."
        sys.exit(0)

    ###########################################################################
    # load all revision map by 'tla ancestry'.
    ###########################################################################
    rmap = RevMap()
    dummyVdic = Vdic()

    if opt_debug == 0:
        rmap.load_ancestry_graph(  map(lambda x: "tla ancestry-graph -m %s |" % x, args)  )
    else:
        rmap.load_ancestry_graph(  args  )

#    rmap.load_ancestry_graph(
#        ["tla ancestry-graph -m lord@emf.net--2003b/tla--devo--1.1 |",
#         "tla ancestry-graph -m stig@brautaset.org--arwen-03/file-utils--stig--1.0 |"])

    rmap.collect_equivalence_class()
#    rmap.dump0()
#    rmap.dump1(1)
    rmap.emit(1, opt_use_symbolic_converter)
    if opt_use_symbolic_converter and g_script_name == '':
        dummyVdic.emit()


if __name__ == '__main__': _main()

# arch-tag: 40a3ee8f-43b7-4b74-abc7-ddaddf2980c8
