#!/usr/bin/python

################################################################
#
# Copyright (C) 2003,2004 Tez Kamihira
#
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
#


#########################################################################
#
# vcache.py -- manipulate chrono chart complex.
#

import os
import vchart
import tlatools
import re
import gauge_rev
import changeset
import sys
import revmap
import vsummary

working_dir = '/w'
control_dir = '/control'
inventory_dir = '/inventory'
lock_file = '/LOCK'
index_d_file = '/dir.map'
index_f_file = '/file.map'
revmap_file = '/revision.map'
vsummary_file = '/vsummary.log'
state_file = '/STATE'
state_tmp_file = '/,,STATE'

def get_revision_no_from_acbvr(vcache_dir, acbvr):

    [a, c, b, v, r] = tlatools.split_acbvr(acbvr)
    if r[:4] == 'dsm-':   # deep survey mode.
        return int(r[4:])
    else:
        rmap = load_revmap(vcache_dir)
        return rmap.finditem(acbvr)

def get_inv_name(vcache_dir, acbvr, f_name):

    f_name = tlatools.trim_dot_slash(f_name)
    cf_fidx = load_index(vcache_dir, 'F')
    r_no = get_revision_no_from_acbvr(vcache_dir, acbvr)
    lines = cf_fidx.check_out_to_list(r_no, 0)

    iv = None
    for i in range(len(lines)):
        m = re.match(r"^(\S+)\s+(\S+)$", lines[i])
        if m != None:
            if tlatools.pika_unescape(m.group(1)) == f_name:
                iv = m.group(2)
                break

    return iv

def load_inv_by_iname(vcache_dir, i_name):

    iv64 = tlatools.iv2iv64(i_name)

    cf_f = vchart.vchart()
    c_path = vcache_dir + inventory_dir + '/' + iv64 + '.gz'
    cf_f.load(c_path)

    return cf_f

def load_inv_by_fname(vcache_dir, acbvr, f_name):

    iv = get_inv_name(vcache_dir, acbvr, f_name)
    if iv == None:
        raise Exception

    return load_inv_by_iname(vcache_dir, iv)

def get_inv_content(vcache_dir, acbvr, i_name, v_cmd):

    cf_f = load_inv_by_iname(vcache_dir, i_name)
    r_no = get_revision_no_from_acbvr(vcache_dir, acbvr)
    lines = cf_f.check_out_to_list(r_no, v_cmd)

    return lines

def load_index(vcache_dir, type):

    cf = vchart.vchart()
    if type == 'F':
        cf.load(vcache_dir + control_dir + index_f_file)
    else:
        cf.load(vcache_dir + control_dir + index_d_file)

    return cf

def load_revmap(vcache_dir):

    rmap = revmap.revmap(vcache_dir + control_dir + revmap_file)
    return rmap

def show_lock_status(super_root):

    if len(super_root) >= 1 and super_root[-1] == '/':
        super_root = super_root[:-1]

    for a in os.listdir(super_root):
        for cbv in os.listdir(super_root + '/' + a):
            if os.access(super_root + '/' + a + '/' + cbv + lock_file, os.F_OK):
                print 'Locked: %s/%s' % (a, cbv)

def kill_locks(super_root):

    if len(super_root) >= 1 and super_root[-1] == '/':
        super_root = super_root[:-1]

    for a in os.listdir(super_root):
        for cbv in os.listdir(super_root + '/' + a):
            kill_file = super_root + '/' + a + '/' + cbv + lock_file
            if os.access(kill_file, os.F_OK):
                os.unlink(kill_file)
                print 'Break lock: %s/%s' % (a, cbv)

def get_state(vcache_dir):

    try:
        if os.access(vcache_dir + state_file, os.F_OK):
            f = open(vcache_dir + state_file)
            l = f.readline()
            f.close()
            m = re.match(r"^(\d+) (\d+)", l)
            if m != None:
                return [ int(m.group(1)), int(m.group(2)) ]
    except Exception:
        pass

    return [0, 0]

def rev_available(vcache_dir, seq):

    [major, minor] = get_state(vcache_dir)
    if minor == 0 and seq < major:
        return 1
    else:
        return 0

def cache_incomplete(vcache_dir):

    [major, minor] = get_state(vcache_dir)
    if minor != 0:
        return 1
    else:
        return 0
    
def get_matched_rev_count(vcache_dir, acbv, b_not_matched = 0):

    rmap = revmap.revmap(vcache_dir + control_dir + revmap_file)
    count = 0
    for i in range(rmap.getsize()):
        if rmap.getitem(i)[ : len(acbv) ] == acbv:
            count = count + 1
    if b_not_matched == 1:
        return rmap.getsize() - count
    else:
        return count

def get_abrowse(vcache_dir):

    v = vsummary.vsummary(vcache_dir + control_dir + vsummary_file)

    #
    # XXX: we can't use 'Revision:' here, because it's not absolute.
    #
    # o = v.get_selective_fields(['Revision:', 'Creator:', 'Standard-date:', 'Summary:'], 0, 9999)

    """ make full qualified revision name with revmap info."""
    rmap = revmap.revmap(vcache_dir + control_dir + revmap_file)
    o = v.get_selective_fields(['Creator:', 'Standard-date:', 'Summary:'], 0, 9999)
    for i in range(len(o)):
        o[i].insert(0, rmap.getitem(i))

    o.reverse() 

    """yank cutting edges info."""
    if len(o) > 0:
        rmap = revmap.revmap(vcache_dir + control_dir + revmap_file)
        acbvr = rmap.getitem(-1)
        [a, c, b, v, r] = tlatools.split_acbvr(acbvr)
        cbv = '%s--%s--%s' % (c, b, v)
        o2 = tlatools.get_abrowse(a, cbv, acbvr)

        # raise Exception, o2
        o = o2 + o

    return o

def is_dsm(vcache_dir):

    rmap = revmap.revmap(vcache_dir + control_dir + revmap_file)
    return rmap.is_dsm()

def v2p_acbvr(vcache_dir, acbvr):

    [a,c,b,v,r] = tlatools.split_acbvr(acbvr)
    rmap = revmap.revmap(vcache_dir + control_dir + revmap_file)

    """XXX: this update timing is still very bogos. :-( """
    rmap.update(a, '%s--%s--%s' % (c,b,v))

    idx = acbvr.rfind('-')
    if idx < 0:
        return "?/?--?--?--?"
    else:
        n = int(acbvr[idx+1:])
        return rmap.getitem(n)

class vcache:

    """Note:

        Some old archive has "C--V" type's version instead of "C--B--V", where
        C, B, V are category, branch, version, respectively.
        Be careful when you split version string by "--" separator.
    """

    def __init__(self, super_root, a, cbv, r = None):

        self.a = a
        self.cbv = cbv
        self.r = r

        self.vcache_dir = '%s/%s/%s' % (super_root, self.a, self.cbv)
        tlatools.mkdir_always(self.vcache_dir + control_dir)
        tlatools.mkdir_always(self.vcache_dir + inventory_dir)

        self.cf_d = None
        self.cf_f = None

        self.major = None
        self.minor = None
        self.get_state()

        self.revmap = revmap.revmap(self.vcache_dir + control_dir + revmap_file)
        if self.revmap.update(self.a, self.cbv) < 0:  # new devline != old line.
            raise Exception

    def load_index(self):

        self.cf_d = load_index(self.vcache_dir, 'D')
        self.cf_f = load_index(self.vcache_dir, 'F')
        self.vsummary = vsummary.vsummary(self.vcache_dir + control_dir + vsummary_file)

    def get_next_vrev_name(self):

        if self.major >= self.revmap.getsize():
            return None
        else:
            return self.revmap.getitem(self.major)

    def get_next_vrev_id(self):

        if self.major >= self.revmap.getsize():
            return None
        else:
            return self.major

    def create_working_space(self):

        tlatools.mkdir_always(self.vcache_dir + working_dir + control_dir)
        tlatools.mkdir_always(self.vcache_dir + working_dir + inventory_dir)

    def remove_working_stuff(self):

        tlatools.rmdir_always(self.vcache_dir + working_dir)
        try:
            os.remove(self.vcache_dir + state_tmp_file)
        except Exception:
            pass

    def move_working_files_to_definitive_area(self):

        tlatools.move_all_files_recursive(
            self.vcache_dir + working_dir + control_dir,
            self.vcache_dir + control_dir)

        tlatools.move_all_files_recursive(
            self.vcache_dir + working_dir + inventory_dir,
            self.vcache_dir + inventory_dir)

    def need_roll_back(self):

        if self.minor == 0:
            if os.access(self.vcache_dir + working_dir, os.F_OK):
                return 1
            else:
                return 0
        else:
            return 0

    def need_roll_forward(self):

        return (self.minor == 1)

    def roll_forward(self):

        tlatools.tmsg('i ROLL FORWARD TO [%s]' % self.revmap.getitem(self.major+1))
        self.move_working_files_to_definitive_area()
        self.remove_working_stuff()
        self.advance_state()

    def roll_back(self):

        tlatools.tmsg('i ROLL BACK TO [%s]' % self.revmap.getitem(self.major))
        self.remove_working_stuff()

    def advance_state(self):

        if self.minor == 0:
            self.set_state(self.major, 1)
        else:
            self.set_state(self.major + 1, 0)

    def get_state(self):

        if self.major != None and self.minor != None:
            return [self.major, self.minor]

        [self.major, self.minor] = get_state(self.vcache_dir)

        return [self.major, self.minor]

    def set_state(self, major, minor):
    
        self.major = major
        self.minor = minor

        f = open(self.vcache_dir + state_tmp_file, 'w')
        f.write('%d %d\n' % (major, minor))
        f.close()

        os.rename(self.vcache_dir + state_tmp_file, self.vcache_dir + state_file)

    def lock(self):

        try:
            f = os.open(self.vcache_dir + lock_file, os.O_CREAT + os.O_EXCL)
        except OSError, val:
            if val[0] == 17:   # file already exists.
                return 0       # lock failed.
            else:
                raise Exception

        os.close(f)
        return 1               # lock succeeded.

    def unlock(self):

        try:
            os.remove(self.vcache_dir + lock_file)
        except Exception:
            pass

    def flush_revmap(self):

        if self.revmap.is_updated():
            self.revmap.save(self.vcache_dir + working_dir + control_dir + revmap_file)

    def get_gauge_revision(self):

        grev = gauge_rev.gauge_rev()
        grev.open(self.revmap.getitem(0), self.vcache_dir + working_dir)
        grev.get_gauge_revision(self.cf_d,
                               self.cf_f,
                               self.vcache_dir + working_dir + inventory_dir)
        b = grev.get_patch_log()
        grev.close()

        self.cf_d.save(self.vcache_dir + working_dir + control_dir + index_d_file)
        self.cf_f.save(self.vcache_dir + working_dir + control_dir + index_f_file)
        self.vsummary.save(self.vcache_dir + working_dir + control_dir + vsummary_file, b)

    def apply_changeset(self):

        tlatools.tmsg('* %s' % self.revmap.getitem(self.major))

        cset = changeset.changeset()
        cset.open('%s' % self.revmap.getitem(self.major))
        cset.apply_changeset(self.cf_f,
                             self.cf_d,
                             self.vcache_dir + inventory_dir,
                             self.vcache_dir + working_dir + inventory_dir,
                             self.major)
        b = cset.get_patch_log()
        cset.close()

        self.cf_d.save(self.vcache_dir + working_dir + control_dir + index_d_file)
        self.cf_f.save(self.vcache_dir + working_dir + control_dir + index_f_file)
        self.vsummary.save(self.vcache_dir + working_dir + control_dir + vsummary_file, b)

def _main():

    itc = vcache("./aaa", "tez@kamihira.com--2003b", "ViewARCH--tez--0.0.1")
    if itc.lock():
        print "lock succeeded"
    else:
        print "lock failed."
        itc.unlock()

if __name__ == '__main__': _main()

# arch-tag: 1a749a74-8b3b-487e-8540-b4bfe6927a84
